# -*- coding: utf-8 -*-
from flask import Flask, request, render_template,json,jsonify,send_from_directory,abort
from StringIO import StringIO
import sys,os
import pandas as pd

import matplotlib
matplotlib.use('Agg')

sys.path.insert(0, './didactic_datamining')

import didactic_datamining as ddm
#from myDaemon import *

app = Flask(__name__)

const = {
    'templates' : {
        'numeric_dataset' : 'numeric_dataset.html',
        'text_dataset': 'text_dataset.html',
		'numeric_dataset_detail' : 'numeric_dataset_detail.html',
		'numeric_dataset_row_template' : 'numeric_dataset_row_template.html',
		'help_page' : 'help.html'
    },
	'numeric_alg_name' : { #devono coincidere con le costanti in JS i nomi degli algoritmi (NUMERIC_ALG_NAME e TEXT_ALG_NAME)
		'KMeans' : 'K-Means',
		'Dbscan' : 'Dbscan',
		'Hierarchical' : 'Hierarchical'
	},
	'text_alg_name' : {
		'Decision_Tree' : 'Decision Tree',
		'Apriori' : 'Apriori',
	}
}

@app.errorhandler(404)
def page_not_found(e):
    return render_template('404.html'), 404

def folderExist(path):
	if not os.path.exists(path):
		os.makedirs(path)

folderExist('./static/output_images')
@app.route('/')
def hello():
	return render_template('index.html',const=const)
	
@app.route('/getTemplate')
def sendTemplate():
	args = request.args.get('args')
	args = json.loads(args) if args else None
	return render_template(request.args.get('templateName'),const=const,args=args)

	
#FUNCTIONS	
#per dataset numerici
def convertNumericDatasetForJS(dataset):
	datasetNew = {}
	c = 0
	for row in dataset:
		datasetNew[c] = [row[0],row[1]]
		c += 1
	return datasetNew

#per dataset numerici
def convertPointsInDataset(points):
	points = json.loads(points) if points else None
	if points is not None:
		x = points['x']
		y = points['y']
		dataset = []
		for c in range(len(x)):
			dataset.append([int(x[c]),int(y[c])])
	return dataset

#per dataset testuale(tabella)
def convertTableInDataset(table):
	return ''



def sendAllImages(imageFileNameInitial,numberOfImages,data):
	filename = request.args.get('filename')
	return render_template('imageAll.html',const=const,filename=imageFileNameInitial,imgN=numberOfImages,data=data)

def sendSpecificImage(imageFileNameInitial,imageNumber,numberOfImages,data):
	filename = request.args.get('filename')
	return render_template('image.html',const=const,filename=imageFileNameInitial,imageNumber=imageNumber,numberOfImages=numberOfImages,data=data)

	
def nestedTemplateToDirectURL(args):
#i templati che vengono innestati dentro la pagina in JS con richieste ajax sono resi accessibili da url dirette attraverso questa funzione
	renderTemplate = ''
	if args and ("embedded" in args) and args["embedded"]:
			if args['algType'] == 'numeric':
				renderTemplate = render_template(const['templates']['numeric_dataset'],const=const,args=args)
			elif args['algType'] == 'text' or args['algType'] == 'text2':
				renderTemplate = render_template(const['templates']['text_dataset'],const=const,args=args)
			elif args['algType'] == 'help':
				renderTemplate = render_template(const['templates']['help_page'],const=const,args=args)
	else:
		renderTemplate = render_template('index.html',const=const,args=args)
	return renderTemplate

@app.route('/Help')
def help():
	args = request.args.get('args')
	args = json.loads(args) if args else {}
	args['algType'] = 'help'
	return nestedTemplateToDirectURL(args)

	
#ALG TEMPLATE	
@app.route('/<algName>')
def chooseAlgTemplate(algName):
	if algName == const['numeric_alg_name']['KMeans']:
		return KMeans(algName)
	elif algName == const['numeric_alg_name']['Dbscan']:
		return Dbscan(algName)
	elif algName == const['numeric_alg_name']['Hierarchical']:
		return Hierarchical(algName)
	elif algName == const['text_alg_name']['Decision_Tree']:
		return DecisionTree(algName)
	elif algName == const['text_alg_name']['Apriori']:
		return Apriori(algName)
	else:
		abort(404)
		
def KMeans(algName):
	args = request.args.get('args')
	args = json.loads(args) if args else {}
	args['algName'] = algName
	args['algType'] = 'numeric'
	return nestedTemplateToDirectURL(args)
	

#@app.route('/Dbscan')
def Dbscan(algName):#tanto condividono lo stesso codice generale
	args = request.args.get('args')
	args = json.loads(args) if args else {}
	args['algName'] = algName
	args['algType'] = 'numeric'
	return nestedTemplateToDirectURL(args)

#@app.route('/Hierarchical')
def Hierarchical(algName):#tanto condividono lo stesso codice generale
	args = request.args.get('args')
	args = json.loads(args) if args else {}
	args['algName'] = algName
	args['algType'] = 'numeric'
	return nestedTemplateToDirectURL(args)
	
#@app.route('/Decision_Tree')
def DecisionTree(algName):#tanto condividono lo stesso codice generale
	args = request.args.get('args')
	args = json.loads(args) if args else {}
	args['algName'] = algName
	args['algType'] = 'text'
	return nestedTemplateToDirectURL(args)	
	
#@app.route('/Apriori')
def Apriori(algName):#tanto condividono lo stesso codice generale
	args = request.args.get('args')
	args = json.loads(args) if args else {}
	args['algName'] = algName
	args['algType'] = 'text2'
	#args['templateName'] = 'index.html'
	return nestedTemplateToDirectURL(args)		
	
#ALG OPTIONS TEMPLATES
@app.route('/<algName>/options')
def chooseAlgOptions(algName):
	if algName == const['numeric_alg_name']['KMeans']:
		return KMeans_options()
	elif algName == const['numeric_alg_name']['Dbscan']:
		return Dbscan_options()
	elif algName == const['numeric_alg_name']['Hierarchical']:
		return Hierarchical_options()
	elif algName == const['text_alg_name']['Decision_Tree']:
		return Decision_Tree_options()
	elif algName == const['text_alg_name']['Apriori']:
		return Apriori_options()
	else:
		abort(404)	
	
#@app.route('/KMeans/options')	
def KMeans_options():
	return render_template('KMeans_options.html')


#@app.route('/Dbscan/options')
def Dbscan_options():
	return render_template('Dbscan_options.html')


#@app.route('/Hierarchical/options')
def Hierarchical_options():
	return render_template('Hierarchical_options.html')

#@app.route('/Decision_Tree/options')
def Decision_Tree_options():
	return render_template('Decision_Tree_options.html')

#@app.route('/Apriori/options')
def Apriori_options():
	return render_template('Apriori_options.html')

	
	
#ALG INFO TEMPLATES
@app.route('/<algName>/info')
def chooseAlgInfoTemplate(algName):
	if algName == const['numeric_alg_name']['KMeans']:
		return KMeans_info()
	elif algName == const['numeric_alg_name']['Dbscan']:
		return Dbscan_info()
	elif algName == const['numeric_alg_name']['Hierarchical']:
		return Hierarchical_info()
	elif algName == const['text_alg_name']['Decision_Tree']:
		return Decision_Tree_info()
	elif algName == const['text_alg_name']['Apriori']:
		return Apriori_info()
	else:
		abort(404)
		
#@app.route('/KMeans/info')
def KMeans_info():
	return render_template('KMeans_info.html')


#@app.route('/Dbscan/info')
def Dbscan_info():
	return render_template('Dbscan_info.html')


#@app.route('/Hierarchical/info')
def Hierarchical_info():
	return render_template('Hierarchical_info.html')

#@app.route('/Decision_Tree/info')
def Decision_Tree_info():
	return render_template('Decision_Tree_info.html')

#@app.route('/Apriori/info')
def Apriori_info():
	return render_template('Apriori_info.html')	




	
#GENERATE DATASET	
@app.route('/<algName>/generate_numeric_dataset')
@app.route('/<algName>/generate_text_dataset')
def chooseDatasetToGenerate(algName):
	if algName == const['numeric_alg_name']['KMeans']:
		return KMeans_generate_numeric_dataset()
	elif algName == const['numeric_alg_name']['Dbscan']:
		return Dbscan_generate_numeric_dataset()
	elif algName == const['numeric_alg_name']['Hierarchical']:
		return Hierarchical_generate_numeric_dataset()
	elif algName == const['text_alg_name']['Decision_Tree']:
		return ""
	elif algName == const['text_alg_name']['Apriori']:
		return Apriori_generate_text_dataset()
	else:
		abort(404)	
		
#@app.route('/KMeans/generate_numeric_dataset')#richiede la creazione di un dataset numerico per Kmeans
def KMeans_generate_numeric_dataset():
	#già dati in js dei valori di default
	algOptions = request.args.get('algOptions')
	algOptions = json.loads(algOptions) if algOptions else None
	
	num_points = int(algOptions['num_points'])
	min_value = int(algOptions['min_value'])
	max_value = int(algOptions['max_value'])
	
	dataset = ddm.create_dataset(npoints=num_points, minvalue=min_value, maxvalue=max_value)
	return jsonify(convertNumericDatasetForJS(dataset))

#@app.route('/Dbscan/generate_numeric_dataset')#richiede la creazione di un dataset numerico per Dbscan
def Dbscan_generate_numeric_dataset():#tanto condividono lo stesso dataset Dbscan, KMeans e Single-Linkage
	return KMeans_generate_numeric_dataset()

#@app.route('/Hierarchical/generate_numeric_dataset')#richiede la creazione di un dataset numerico per Hierarchical
def Hierarchical_generate_numeric_dataset():#tanto condividono lo stesso dataset Dbscan, KMeans e Single-Linkage
	return KMeans_generate_numeric_dataset()

#@app.route('/Apriori/generate_text_dataset')#richiede la creazione di un dataset testuale per Apriori
def Apriori_generate_text_dataset():
	transactions = ddm.create_transactional_dataset(num_transaction=10, num_items=6, min_len=2, max_len=4)
	return jsonify(transactions)
	
	
#DRAW DATASET	
@app.route('/<algName>/draw_dataset')
def chooseDatasetToDraw(algName):
	if algName == const['numeric_alg_name']['KMeans']:
		return draw_dataset_KMeans()
	elif algName == const['numeric_alg_name']['Dbscan']:
		return draw_dataset_Dbscan()
	elif algName == const['numeric_alg_name']['Hierarchical']:
		return draw_dataset_Hierarchical()
	elif algName == const['text_alg_name']['Decision_Tree']:
		return draw_dataset_Decision_Tree()
	elif algName == const['text_alg_name']['Apriori']:
		return draw_dataset_Apriori()
	else:
		abort(404)	
		
#@app.route('/KMeans/draw_dataset')#serve a disegnare un dataset
def draw_dataset_KMeans():
	points = request.args.get('points')
	dataset = convertPointsInDataset(points)
	imageNumber = int(request.args.get('imageNumber'))
	stepByStep = False if request.args.get('stepByStep') == 'false' else True
	algOptions = request.args.get('algOptions')
	algOptions = json.loads(algOptions) if algOptions else None
	
	#gia' dati in JS dei valori di default
	centroid_index_c1 = int(algOptions['centroid_index_c1'])
	centroid_index_c2 = int(algOptions['centroid_index_c2'])
		
	if len(dataset) > 0:

		kmeans = ddm.DidatticKMeans(K=2, centroid_indexs=(centroid_index_c1, centroid_index_c2), dist=ddm.euclidean_distance)
		imageFileNameInitial,numberOfImages = kmeans.fit(dataset, step_by_step=False)
#		print "numero immagini",numberOfImages
		if stepByStep:
			return sendSpecificImage(imageFileNameInitial,imageNumber,numberOfImages,{})
		else:
			return sendAllImages(imageFileNameInitial,numberOfImages,{})

			
			
#@app.route('/Dbscan/draw_dataset')
def draw_dataset_Dbscan():
	points = request.args.get('points')
	dataset = convertPointsInDataset(points)
	imageNumber = int(request.args.get('imageNumber'))
	stepByStep = False if request.args.get('stepByStep') == 'false' else True
	algOptions = request.args.get('algOptions')
	algOptions = json.loads(algOptions) if algOptions else None
	
	#gia' dati in JS dei valori di default
	eps = float(algOptions['eps'])
	min_pts = int(algOptions['min_pts'])
	
	if len(dataset) > 0:
		dbscan = ddm.DidatticDbscan(eps=eps, min_pts=min_pts)
		imageFileNameInitial,numberOfImages,data = dbscan.fit(dataset, step_by_step=False)
#		print "numero immagini",numberOfImages
		data['algName'] = const['numeric_alg_name']['Dbscan']
		if stepByStep:
			return sendSpecificImage(imageFileNameInitial,imageNumber,numberOfImages,data)
		else:
			return sendAllImages(imageFileNameInitial,numberOfImages,data)

			
			
#@app.route('/Hierarchical/draw_dataset')
def draw_dataset_Hierarchical():
	points = request.args.get('points')
	dataset = convertPointsInDataset(points)
	imageNumber = int(request.args.get('imageNumber'))
	stepByStep = False if request.args.get('stepByStep') == 'false' else True
	algOptions = request.args.get('algOptions')
	algOptions = json.loads(algOptions) if algOptions else None
	
	#gia' dati in JS dei valori di default
	link_criteria = algOptions['link_criteria']
	use_distances = True if algOptions['use_distances'] == 'true' else False
	if len(dataset) > 0:
		hier = ddm.DidatticHierarchical()
		imageFileNameInitial,numberOfImages,data = hier.fit(dataset, link_criteria=link_criteria, use_distances=use_distances, step_by_step=False)
		data['algName'] = const['numeric_alg_name']['Hierarchical']
		if stepByStep:
			return sendSpecificImage(imageFileNameInitial,imageNumber,numberOfImages,data)
		else:
			return sendAllImages(imageFileNameInitial,numberOfImages,data)


#@app.route('/Decision_Tree/draw_dataset')
def draw_dataset_Decision_Tree():
	#table = request.args.get('points')
	algOptions = request.args.get('algOptions')
	algOptions = json.loads(algOptions) if algOptions else None
	
	#gia' dati in JS dei valori di default
	fun_name = algOptions['fun_name']
	min_samples_split = algOptions['min_samples_split']
	min_samples_leaf = algOptions['min_samples_leaf']
	target = request.args.get('target')
	text = request.args.get('text')
	text_old_dataset = request.args.get('oldDataset')
	predict = True if request.args.get('predict') == 'true' else False
	if len(text) > 0:
		if fun_name == 'error':
			tree = ddm.DidatticClassificationTree(fun=ddm.error_rate, fun_name='error_rate', 
												  min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf, step_by_step=False)
		if fun_name == 'gini':
			tree = ddm.DidatticClassificationTree(fun=ddm.gini, fun_name='gini', 
												  min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf, step_by_step=False)
		
		if not predict:
			dataset_df = pd.read_csv(StringIO(text),skipinitialspace=True,delimiter=',')
			calculus_string,imageFileNameInitial,evaluation_string = tree.fit(dataset_df, target=target)
			return render_template('Decision_Tree_ImageAll.html',const=const,filename=imageFileNameInitial,calculus_string=calculus_string)
		else:
			dataset_df = pd.read_csv(StringIO(text_old_dataset),skipinitialspace=True,delimiter=',')
			calculus_string,imageFileNameInitial,evaluation_string = tree.fit(dataset_df, target=target)
			dataset_df = pd.read_csv(StringIO(text),skipinitialspace=True,delimiter=',')
			prediction = tree.predict(dataset_df)
			dataset_df['Predicted'] = prediction
			evaluation_string=tree.evaluate(dataset_df);
#			print evaluation_string
			evaluation_string = evaluation_string.replace('\t','',1)
#			print evaluation_string
			return render_template('Decision_Tree_ImageAll.html',const=const,filename=imageFileNameInitial,calculus_string=calculus_string,predict=evaluation_string,dataset_df=dataset_df)


#@app.route('/Apriori/draw_dataset')
def draw_dataset_Apriori():
	#table = request.args.get('points')
	algOptions = request.args.get('algOptions')
	algOptions = json.loads(algOptions) if algOptions else None
	
	#gia' dati in JS dei valori di default
	min_sup = algOptions['min_sup']
	sup_type = algOptions['sup_type']
	min_conf = algOptions['min_conf']
	text = request.args.get('text')
	points = request.args.get('points')
	text_old_dataset = request.args.get('oldDataset')
	predict = True if request.args.get('predict') == 'true' else False
	if len(text) > 0:

		transactions = json.loads(points)
		apriori = ddm.DidatticApriori(min_sup=float(min_sup), sup_type=sup_type)
		self,data = apriori.fit(transactions, step_by_step=False)
		data["rules"] = apriori.extract_rules(min_conf=float(min_conf))
#		print data
		data['algName'] = const['text_alg_name']['Apriori']
		return render_template('Apriori_ImageAll.html',const=const,data=data)



if __name__ == '__main__':
	app.run(host='0.0.0.0', debug=True)
